import os, shutil
import config
from tqdm import tqdm
import numpy as np
import matplotlib.pyplot as plt
import mne
import scipy
from scipy import signal, stats

def BinarizeChannels(network_channels=config.network_channels):
    bin_str = ""
    for channel in config.channel_names:
        if channel in network_channels:
            bin_str = bin_str + "1"
        else:
            bin_str = bin_str + "0"
    return(bin_str)

# just keeps subset of 19 channels as defined in network_channels
def FilterChannels(array, keep_channels, axisNum):
    filter_indeces = []
    # print("Old Shape of Dataset:", dataset.shape, "\n")
    # for each channel in my channel list
    for keep in keep_channels:
        filter_indeces.append(config.channel_names.index(keep))
    #   get the index of that channel in the master channel list
    #   add that index number to a new list filter_indeces

    newarr = np.take(array, filter_indeces, axisNum)
    # print("New Shape of Dataset:", filtered_dataset.shape, "\n")
    return(newarr)

def generate_sparse_indexes(art, length):
    i = 0
    startindexes = []
    while i < art.shape[0]-length:
        stk = art[i:(i+length),:]
        if not np.any(np.isnan(stk)):
            startindexes.append(i)
            i+=length
        else:
            i+=1
    return startindexes

# takes one positional argument, path of study folder
class StudyFolder:

    def __new__(self, path):
        if os.path.isdir(path):
            return super(StudyFolder, self).__new__(self)
        else:
            print("The path supplied is not a valid directory.")
            raise ValueError

    def __init__(self, path):
        self.path = path
        print("\nInitializing New Study Directory at: " + self.path)
        self.raw_fnames = os.listdir(self.path+"/raw")

    # standardizes task names / file structure and anonymizes subject headers
    # translator stored in StudyFolder/task_translator.txt
    def autoclean(self):
        for task in config.tasks:
            for irregular in config.tasks[task]:
                self.standardize(irregular, task)
            if os.path.isdir(self.path+"/"+task):
                self.anon(task)
        if len(os.listdir(self.path)) != 0:
            print("Some raw files couldn't be automatically standardized. You should review them in /raw before moving forward with analysis.")

    def set_raw_fnames(self):
        self.raw_fnames = os.listdir(self.path+"/raw")

    def get_task_fnames(self, task):
        return(os.listdir(self.path+"/"+task))

    # standardizes every filename possible, using alternative (unclean)
    # fnames from the WAVi desktop which are written in the tasks dict in config.py
    def standardize(self, old, new):
        if not os.path.isdir(self.path+"/"+new):
            print("Making new task folder: ", new)
            os.mkdir(self.path+"/"+new)

        for fname in [fname for fname in self.raw_fnames if old in fname]:
            newfname = fname.replace(old, new)
            shutil.move(self.path+"/raw/"+fname, self.path+"/"+new+"/"+newfname)

        self.set_raw_fnames()
        if len(os.listdir(self.path+"/"+new))==0:
            os.rmdir(self.path+"/"+new)

    # anonymizes sets of standardized task data which can then be read into a Trials object
    def anon(self, task, groupNum=1):
        translator = {}
        subject_leads = set([fname.replace(task, '')[:-4] for fname in self.get_task_fnames(task)])
        i = 0
        f = open(self.path+"/translator_"+task+".txt", "w")
        for lead in subject_leads:
            translator[lead] = str(groupNum)+"0"*(config.participantNumLen-len(str(i))-1)+str(i)
            i+=1
            f.write(lead)
            f.write("\t")
            f.write(translator[lead])
            f.write("\n")
            files = [fname for fname in self.get_task_fnames(task) if lead in fname]
            for file in files:
                newfile = file.replace(lead, translator[lead]+"_")
                shutil.move(self.path+"/"+task+"/"+file, self.path+"/"+task+"/"+newfile)


# takes one positional argument, path of trials folder
class Trials:

    def __new__(self, path):
        if os.path.isdir(path):
            return super(Trials, self).__new__(self)
        else:
            print("The path supplied is not a valid directory.")
            raise ValueError

    def __init__(self, path):
        self.path = path
        self.studyFolder = os.path.dirname(path)
        self.task = os.path.basename(self.path)
        self.task_fnames = os.listdir(self.path)
        self.network_channels = config.network_channels

    def get_task_fnames(self, task):
        return(os.listdir(self.path))

    def no_filter_rename(self):
        for fname in self.task_fnames:
            shutil.move(self.path+"/"+fname, self.path+"/"+fname[:-4]+"_nofilter"+fname[-4:])
        self.task_fnames = self.get_task_fnames(self.task)

    def set_subjects(self):
        self.subjects = set([fname[:config.participantNumLen] for fname in self.get_task_fnames(self.task)])

    # takes length (in samples @ 250 Hz / config.sampleRate) as positional argument
    def gen_contigs(self, contigLength, network_channels=config.network_channels, artDegree=config.artDegree):

        if not hasattr(self, 'subjects'):
            self.set_subjects()

        if not hasattr(self, 'contigs'):
            self.contigs = []

        # make parent contigs folder
        if not os.path.isdir(self.studyFolder+"/contigs"):
            os.mkdir(self.studyFolder+"/contigs")

        # make a child subdirectory called contigs_<task>_<contig_length>
        try:
            self.contigsFolder = self.studyFolder+"/contigs/"+self.task+"_"+str(contigLength)+"_"+BinarizeChannels(network_channels=self.network_channels)+"_"+str(artDegree)
            os.mkdir(self.contigsFolder)
        except:
            print("Couldn't create the specified contigs folder.\n")

        print("Contigifying Data:\n====================")

        for sub in tqdm(self.subjects):
            artfile = sub+"_"+self.task+"_nofilter.art"

            # load in artifact file as np array
            # print("Artifact:"+self.path+"/"+artfile)
            artifact = np.genfromtxt(self.path+"/"+artfile, delimiter=" ")

            if artifact.size == 0:
                print("Most likely an empty text file was encountered. Skipping: " + artfile)
                continue

            else:
                # get rid of channels we don't want the net to use
                artifact = FilterChannels(artifact, network_channels, 1)
                # mask artifact array where numbers exceed artDegree
                mxi = np.ma.masked_where(artifact > artDegree, artifact)
                mxi = np.ma.filled(mxi.astype(float), np.nan)
                artifact = mxi

                # write list of start indexes for windows which meet
                # contig requirements
                i = 0
                indeces = []
                while i < artifact.shape[0]-contigLength:
                    stk = artifact[i:(i+contigLength),:]
                    if not np.any(np.isnan(stk)):
                        indeces.append(i)
                        i+=contigLength
                    else:
                        i+=1

                subfiles = [fname for fname in self.task_fnames if (sub==fname[:config.participantNumLen] and "eeg" in fname)]
                # subfiles = [fname for fname in subfiles if "eeg" in fname]

                j = 0
                for eegfile in subfiles:
                    # print("EEG file:"+self.path+"/"+eegfile)
                    data = np.genfromtxt(self.path+"/"+eegfile, delimiter=" ")
                    if data.size == 0:
                        print("Most likely an empty text file was encountered. Skipping: " + eegfile)

                    else:
                        band = eegfile.split('_')[2][:-4]
                        for i in indeces:
                            self.contigs.append(Contig(data[i:(i+contigLength),:], i, sub, band))

    def write_contigs(self):
        if not hasattr(self, 'contigs'):
            print("This instance of the class 'Trials' has no 'contigs' attribute.")
            raise ValueError
        else:
            print("Writing Contigs:\n====================")
            for contig in tqdm(self.contigs):
                contig.write(self.contigsFolder+"/"+contig.subject+"_"+contig.band+"_"+str(contig.startindex)+".csv")

    # takes length (in samples @ 250 Hz) as positional argument
    def gen_spectra(self, contigLength, network_channels=config.network_channels, artDegree=config.artDegree):
        if not hasattr(self, 'spectra'):
            self.spectra = []
        self.contigsFolder = self.studyFolder+"/contigs/"+self.task+"_"+str(contigLength)+"_"+BinarizeChannels(network_channels=self.network_channels)+"_"+str(artDegree)
        self.spectraFolder = self.studyFolder+"/spectra/"+self.task+"_"+str(contigLength)+"_"+BinarizeChannels(network_channels=self.network_channels)+"_"+str(artDegree)
        # make parent spectra folder
        if not os.path.isdir(self.studyFolder+"/spectra"):
            os.mkdir(self.studyFolder+"/spectra")

        # make a child subdirectory called <task>_<contig_length>_<binary_channels_code>
        try:
            self.spectraFolder = self.studyFolder+"/spectra/"+self.task+"_"+str(contigLength)+"_"+BinarizeChannels(network_channels=self.network_channels)+"_"+str(artDegree)
            os.mkdir(self.spectraFolder)
        except:
            print("Couldn't create the specified spectra folder.\n")

        print("Fourier Transforming Data:\n====================")
        for contig in tqdm(os.listdir(self.contigsFolder)):
            self.spectra.append(Contig(np.genfromtxt(self.contigsFolder+"/"+contig, delimiter=","), contig.split('_')[2][:-4], contig[:config.participantNumLen], contig.split('_')[1]).fft())

    def write_spectra(self):
        if not hasattr(self, 'spectra'):
            print("This instance of the class 'Trials' has no 'spectra' attribute.")
            raise ValueError
        else:
            print("Writing Spectra:\n====================")
            for spectrum in tqdm(self.spectra):
                spectrum.write(self.spectraFolder+"/"+spectrum.subject+"_"+spectrum.band+"_"+str(spectrum.startindex)+".csv")

# takes 4 positional arguments: data array, sample of window start, subject ID, and frequency band or "nofilter"
class Contig:
    def __init__(self, data, startindex, subject, band, source=None):
        self.data = data
        self.startindex = startindex
        self.subject = subject
        self.band = band
        self.group = self.subject[0]
        self.source = source

    def write(self, path):
        np.savetxt(path, self.data, delimiter=",", fmt="%2.1f")

    def fft(self):
        channel_number = 0
        # spectral = np.ndarray(shape=(len(self.data), len(config.channel_names))).T
        for sig in self.data.T:
            electrode = config.network_channels[channel_number]
            # perform pwelch routine to extract PSD estimates by channel
            f, Pxx_den = scipy.signal.periodogram(
                sig,
                fs=float(config.sampleRate),
                window='hann'
                )
            if channel_number == 0:
                spectral = np.array(f)
            spectral = np.column_stack((spectral, Pxx_den))
            channel_number+=1
        return(Spectra(spectral, spectral.T[0], self.startindex, self.subject, self.band))

    def plot(self):
        fig, axs = plt.subplots(nrows=1, figsize=(16, 8))

        trials = []
        plots = []
        for channel in config.network_channels:
            i = config.network_channels.index(channel)
            trial = self.data.T[i]
            trials.append(trial)
            wavs = trial * 0.1
            print(wavs)
            t = np.arange(0, len(wavs)) / config.sampleRate
            wavs = np.subtract(wavs, i*25)
            plots.append(axs.plot(t, wavs)[0])

        axs.axis([t[0], t[-1], -25*len(config.network_channels), 5*len(config.network_channels)])
        leg = axs.legend(config.network_channels, loc="right", fancybox=True)
        leg.get_frame().set_alpha(0.4)
        axs.grid(b=True)
        axs.set_title("Raw Waveforms: Subject " + self.subject)
        axs.set_ylabel("Voltage [mV]")
        axs.set_xlabel('Time [seconds]')

        # we will set up a dict mapping legend line to orig line, and enable
        # picking on the legend line
        lined = dict()
        for legline, origline in zip(leg.get_lines(), plots):
            legline.set_picker(5)  # 5 pts tolerance
            lined[legline] = origline

        def onpick(event):
            # on the pick event, find the orig line corresponding to the
            # legend proxy line, and toggle the visibility
            legline = event.artist
            origline = lined[legline]
            vis = not origline.get_visible()
            origline.set_visible(vis)
            # Change the alpha on the line in the legend so we can see what lines
            # have been toggled
            if vis:
                legline.set_alpha(1.0)
            else:
                legline.set_alpha(0.2)
            fig.canvas.draw()

        fig.canvas.mpl_connect('pick_event', onpick)

        def movonpick(event):
            # on the pick event, find the orig line corresponding to the
            # legend proxy line, and toggle the visibility
            movlegline = event.artist
            movorigline = movlined[movlegline]
            vis = not movorigline.get_visible()
            movorigline.set_visible(vis)
            # Change the alpha on the line in the legend so we can see what lines
            # have been toggled
            if vis:
                movlegline.set_alpha(1.0)
            else:
                movlegline.set_alpha(0.2)
            fig.canvas.draw()

        fig.canvas.mpl_connect('pick_event', movonpick)

        plt.tight_layout()
        plt.show()

class Spectra:
    def __init__(self, data, freq, startindex, subject, band="nofilter", channels=config.network_channels, source=None):
        self.data = data
        self.freq = freq
        self.startindex = startindex
        self.subject = subject
        self.band = band
        self.group = int(self.subject[0])
        self.channels = channels
        self.freq_res = (config.sampleRate/2) / (len(self.freq)-1)
        self.contig_length = config.sampleRate // self.freq_res
        self.source = source

    def write(self, path):
        np.savetxt(path, self.data, delimiter=",", fmt="%2.1f")



import Prep, os
import config
import numpy as np
import matplotlib.pyplot as plt

# takes one positional argument, object from ML.Classifier
class SpectralAverage:
    def __new__(self, inputClf):
        if inputClf.type == "spectra":
            return super(SpectralAverage, self).__new__(self)
        else:
            print("This class requires an ML.Classifier object with type 'spectra' as a positional argument.")
            raise ValueError

    def __init__(self, inputClf):
        self.Clf = inputClf
        self.f = self.Clf.data[0].freq
        self.avgs = []
        self.groups = list(set(SpecObj.group for SpecObj in self.Clf.data))
        for group in self.groups:
            dataset = np.stack([np.sqrt(SpecObj.data) for SpecObj in self.Clf.data if SpecObj.group == group])
            self.avgs.append(np.mean(dataset, axis=0))

    def plot(self, channels=config.network_channels):
        plt.rcParams['figure.dpi'] = 200
        num_waves = len(self.avgs)
        num_channels = self.avgs[0].shape[-1]
        channels = channels

        fig, axs = plt.subplots(nrows=len(channels), sharex=False, figsize=(20, 40))
        fig.tight_layout()

        axs[-1].set_xlabel('Frequency')

        i = 0
        for channel in channels:
            j = 0
            for array in self.avgs:
                axs[i].plot(self.f[:100], array.T[i][:100], label=self.groups[j])
                j+=1
            axs[i].set_title(channel)
            axs[i].set_ylabel('Magnitude')
            axs[i].legend()
            i+=1

        plt.show()



import os
import numpy as np
import config
# import scipy
import random
import math
from tqdm import tqdm
import sklearn
from sklearn.svm import LinearSVC
from sklearn.svm import SVC
from sklearn.model_selection import train_test_split
from sklearn import svm, metrics
from sklearn.preprocessing import StandardScaler
from sklearn.preprocessing import MinMaxScaler
from sklearn.pipeline import make_pipeline
from sklearn.model_selection import StratifiedShuffleSplit
from sklearn.model_selection import GridSearchCV
from sklearn.metrics import average_precision_score
from sklearn.metrics import precision_recall_curve
from sklearn.metrics import plot_precision_recall_curve
from sklearn.feature_selection import SelectKBest, f_classif
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize

from Prep import Contig, Spectra

# Utility function to move the midpoint of a colormap to be around
# the values of interest.
class MidpointNormalize(Normalize):

    def __init__(self, vmin=None, vmax=None, midpoint=None, clip=False):
        self.midpoint = midpoint
        Normalize.__init__(self, vmin, vmax, clip)

    def __call__(self, value, clip=None):
        x, y = [self.vmin, self.midpoint, self.vmax], [0, 0.5, 1]
        return np.ma.masked_array(np.interp(value, x, y))

# class ZscoreStandardize(normal):
#
#     def __init__(self, vmin=None, vmax=None, mean=None, clip=False):
#         self.mean = mean
#

# takes one positional argument: either "contigs" or "spectra"
class Classifier:
    def __init__(self, type, network_channels=config.network_channels):
        self.data = []
        self.type = type
        self.network_channels=config.network_channels
        self.trial_name = None

    def LoadData(self, path):
        fname = os.path.basename(path)
        self.trial_name = os.path.basename(os.path.dirname(os.path.dirname(path)))+"/"+os.path.basename(os.path.dirname(path))
        if self.type == "contigs":
            self.data.append(Contig(np.genfromtxt(path, delimiter=","), fname.split('_')[2][:-4], fname[:config.participantNumLen], fname.split('_')[1]))
        elif self.type == "spectra":
            self.data.append(Spectra(np.genfromtxt(path, delimiter=",").T[1:].T, np.genfromtxt(path, delimiter=",").T[0].T, fname.split('_')[2][:-4], fname[:config.participantNumLen], fname.split('_')[1], source=os.path.basename(os.path.dirname(path))))

    def Balance(self, parentPath):
        folders = ["ref 24-30", "ref 31-40", "ref 41-50", "ref 51-60", "ref 61-70", "ref 71-80", "ref 81+"]
        spectralFolder = self.trial_name
        totalpos = 0
        totalneg = 0

        for item in self.data:
            if item.group == 1:
                totalneg+=1
            elif item.group == 2:
                totalpos+=1

        fill = (totalpos - totalneg) / len(folders)
        for folder in tqdm(folders):
            fnames = os.listdir(parentPath+"/"+folder+"/"+spectralFolder)
            random.shuffle(fnames)
            i = 0
            while i < fill:
                self.LoadData(parentPath+"/"+folder+"/"+spectralFolder+"/"+fnames[i])
                i+=1

    def LDA(self, normalize='standard', tt_split=0.33, lowbound=3, highbound=20, plot_data=False):
        # shuffle dataset
        random.shuffle(self.data)

        # Features = np.array(self.data[0].freq[int(lowbound//self.data[0].freq_res)+1:int(highbound//self.data[0].freq_res)+2])

        # print("Shape of one data:", self.data[0].data.shape)

        # total num for each class
        totalpos = 0
        totalneg = 0

        for item in self.data:
            if item.group == 1:
                totalneg+=1
            elif item.group == 2:
                totalpos+=1

        print("Number of negative outcomes:", totalneg)
        print("Number of positive outcomes:", totalpos)

        # make list of subjects in this dataset
        subjects = list(set([(item.source, item.subject) for item in self.data]))

        # shuffle subject list randomly and then
        # split data into train and test spectra (no overlapping subjects)
        # "subject-stratified" train-test split
        random.shuffle(subjects)
        split = math.floor(len(subjects)*tt_split)
        train_subjects = subjects[:split*-1]
        test_subjects = subjects[split*-1:]

        if normalize == 'standard':
            myscaler = StandardScaler()
        elif normalize == 'minmax':
            myscaler = MinMaxScaler()
        else:
            normalize = None

        if self.type == "spectra":
            # load dataset into np arrays / train & test
            # make label maps for each set: 1 = positive condition, -1 = negative condition
            train_dataset = np.stack([SpecObj.data[int(lowbound//SpecObj.freq_res)+1:int(highbound//SpecObj.freq_res)+2] for SpecObj in self.data if (SpecObj.source, SpecObj.subject) in train_subjects])
            train_labels = np.array([int(1) if (SpecObj.group==2) else int(-1) for SpecObj in self.data if (SpecObj.source, SpecObj.subject) in train_subjects])
            # print("Number of train examples:", len(train_dataset))


            test_dataset = np.stack([SpecObj.data[int(lowbound//SpecObj.freq_res)+1:int(highbound//SpecObj.freq_res)+2] for SpecObj in self.data if (SpecObj.source, SpecObj.subject) in test_subjects])
            test_labels = np.array([int(1) if (SpecObj.group==2) else int(-1) for SpecObj in self.data if (SpecObj.source, SpecObj.subject) in test_subjects])
            # print("Number of test examples:", len(test_dataset))

            # print("Frequency Resolution:", self.data[0].freq_res, "Hz")

        # reshape datasets
        nsamples, nx, ny = train_dataset.shape
        train_dataset = train_dataset.reshape((nsamples, nx*ny))

        nsamples, nx, ny = test_dataset.shape
        test_dataset = test_dataset.reshape((nsamples, nx*ny))

        # << Normalize features (z-score standardization) >> #
        # create classifier pipeline
        if normalize != None:
            clf = make_pipeline(myscaler, LinearDiscriminantAnalysis())
        else:
            clf = make_pipeline(LinearDiscriminantAnalysis())

        # fit to train data
        clf.fit(train_dataset, train_labels)
        print('Classification accuracy on validation data: {:.3f}'.format(clf.score(test_dataset, test_labels)))

        if plot_data == True:
            # #############################################################################
            # Colormap
            cmap = colors.LinearSegmentedColormap(
                'red_blue_classes',
                {'red': [(0, 1, 1), (1, 0.7, 0.7)],
                 'green': [(0, 0.7, 0.7), (1, 0.7, 0.7)],
                 'blue': [(0, 0.7, 0.7), (1, 1, 1)]})
            plt.cm.register_cmap(cmap=cmap)


            # #############################################################################
            # Generate datasets
            def dataset_fixed_cov():
                '''Generate 2 Gaussians samples with the same covariance matrix'''
                n, dim = 300, 2
                np.random.seed(0)
                C = np.array([[0., -0.23], [0.83, .23]])
                X = np.r_[np.dot(np.random.randn(n, dim), C),
                          np.dot(np.random.randn(n, dim), C) + np.array([1, 1])]
                y = np.hstack((np.zeros(n), np.ones(n)))
                return X, y


            def dataset_cov():
                '''Generate 2 Gaussians samples with different covariance matrices'''
                n, dim = 300, 2
                np.random.seed(0)
                C = np.array([[0., -1.], [2.5, .7]]) * 2.
                X = np.r_[np.dot(np.random.randn(n, dim), C),
                          np.dot(np.random.randn(n, dim), C.T) + np.array([1, 4])]
                y = np.hstack((np.zeros(n), np.ones(n)))
                return X, y


            # #############################################################################
            # Plot functions
            def plot_data(lda, X, y, y_pred, fig_index):
                splot = plt.subplot(2, 2, fig_index)
                if fig_index == 1:
                    plt.title('Linear Discriminant Analysis')
                    plt.ylabel('Data with\n fixed covariance')
                elif fig_index == 2:
                    plt.title('Quadratic Discriminant Analysis')
                elif fig_index == 3:
                    plt.ylabel('Data with\n varying covariances')

                tp = (y == y_pred)  # True Positive
                tp0, tp1 = tp[y == 0], tp[y == 1]
                X0, X1 = X[y == 0], X[y == 1]
                X0_tp, X0_fp = X0[tp0], X0[~tp0]
                X1_tp, X1_fp = X1[tp1], X1[~tp1]

                # class 0: dots
                plt.scatter(X0_tp[:, 0], X0_tp[:, 1], marker='.', color='red')
                plt.scatter(X0_fp[:, 0], X0_fp[:, 1], marker='x',
                            s=20, color='#990000')  # dark red

                # class 1: dots
                plt.scatter(X1_tp[:, 0], X1_tp[:, 1], marker='.', color='blue')
                plt.scatter(X1_fp[:, 0], X1_fp[:, 1], marker='x',
                            s=20, color='#000099')  # dark blue

                # class 0 and 1 : areas
                nx, ny = 200, 100
                x_min, x_max = plt.xlim()
                y_min, y_max = plt.ylim()
                xx, yy = np.meshgrid(np.linspace(x_min, x_max, nx),
                                     np.linspace(y_min, y_max, ny))
                Z = lda.predict_proba(np.c_[xx.ravel(), yy.ravel()])
                Z = Z[:, 1].reshape(xx.shape)
                plt.pcolormesh(xx, yy, Z, cmap='red_blue_classes',
                               norm=colors.Normalize(0., 1.), zorder=0)
                plt.contour(xx, yy, Z, [0.5], linewidths=2., colors='white')

                # means
                plt.plot(lda.means_[0][0], lda.means_[0][1],
                         '*', color='yellow', markersize=15, markeredgecolor='grey')
                plt.plot(lda.means_[1][0], lda.means_[1][1],
                         '*', color='yellow', markersize=15, markeredgecolor='grey')

                return splot


            def plot_ellipse(splot, mean, cov, color):
                v, w = linalg.eigh(cov)
                u = w[0] / linalg.norm(w[0])
                angle = np.arctan(u[1] / u[0])
                angle = 180 * angle / np.pi  # convert to degrees
                # filled Gaussian at 2 standard deviation
                ell = mpl.patches.Ellipse(mean, 2 * v[0] ** 0.5, 2 * v[1] ** 0.5,
                                          180 + angle, facecolor=color,
                                          edgecolor='black', linewidth=2)
                ell.set_clip_box(splot.bbox)
                ell.set_alpha(0.2)
                splot.add_artist(ell)
                splot.set_xticks(())
                splot.set_yticks(())


            def plot_lda_cov(lda, splot):
                plot_ellipse(splot, lda.means_[0], lda.covariance_, 'red')
                plot_ellipse(splot, lda.means_[1], lda.covariance_, 'blue')


            def plot_qda_cov(qda, splot):
                plot_ellipse(splot, qda.means_[0], qda.covariance_[0], 'red')
                plot_ellipse(splot, qda.means_[1], qda.covariance_[1], 'blue')

            plt.figure(figsize=(10, 8), facecolor='white')
            plt.suptitle('Linear Discriminant Analysis vs Quadratic Discriminant Analysis',
                         y=0.98, fontsize=15)
            for i, (X, y) in enumerate([dataset_fixed_cov(), dataset_cov()]):
                # Linear Discriminant Analysis
                lda = LinearDiscriminantAnalysis(solver="svd", store_covariance=True)
                y_pred = lda.fit(X, y).predict(X)
                splot = plot_data(lda, X, y, y_pred, fig_index=2 * i + 1)
                plot_lda_cov(lda, splot)
                plt.axis('tight')

                # Quadratic Discriminant Analysis
                qda = QuadraticDiscriminantAnalysis(store_covariance=True)
                y_pred = qda.fit(X, y).predict(X)
                splot = plot_data(qda, X, y, y_pred, fig_index=2 * i + 2)
                plot_qda_cov(qda, splot)
                plt.axis('tight')
            plt.tight_layout()
            plt.subplots_adjust(top=0.92)
            plt.show()


    def SVM(self, C=1, kernel=config.kernel_type, normalize=None, iterations=1000, plot_PR=False, plot_Features=False, feat_select=False, num_feats=10, lowbound=3, highbound=20, tt_split=0.33):
        # shuffle dataset
        random.shuffle(self.data)

        Features = np.array(self.data[0].freq[int(lowbound//self.data[0].freq_res)+1:int(highbound//self.data[0].freq_res)+2])

        # print("Shape of one data:", self.data[0].data.shape)

        # total num for each class
        totalpos = 0
        totalneg = 0

        for item in self.data:
            if item.group == 1:
                totalneg+=1
            elif item.group == 2:
                totalpos+=1

        print("Number of negative outcomes:", totalneg)
        print("Number of positive outcomes:", totalpos)


        # make list of subjects in this dataset
        subjects = list(set([(item.source, item.subject) for item in self.data]))


        # << old split methods >> #
                    # split data into train and test spectra (not split by subjects)
                    # x_train, x_test, y_train, y_test = train_test_split(dataset, labelmap, test_size=0.3)

                    # split data into train and test spectra (stratified by subjects)
                    # x_train, x_test, y_train, y_test = train_test_split(dataset, labelmap, test_size=0.3)


        # shuffle subject list randomly and then
        # split data into train and test spectra (no overlapping subjects)
        # "subject-stratified" train-test split
        random.shuffle(subjects)
        split = math.floor(len(subjects)*tt_split)
        train_subjects = subjects[:split*-1]
        test_subjects = subjects[split*-1:]

        if normalize == 'standard':
            myscaler = StandardScaler()
        elif normalize == 'minmax':
            myscaler = MinMaxScaler()
        else:
            normalize = None

        if self.type == "spectra":
            # load dataset into np arrays / train & test
            # make label maps for each set: 1 = positive condition, -1 = negative condition
            train_dataset = np.stack([SpecObj.data[int(lowbound//SpecObj.freq_res)+1:int(highbound//SpecObj.freq_res)+2] for SpecObj in self.data if (SpecObj.source, SpecObj.subject) in train_subjects])
            train_labels = np.array([int(1) if (SpecObj.group==2) else int(-1) for SpecObj in self.data if (SpecObj.source, SpecObj.subject) in train_subjects])
            # print("Number of train examples:", len(train_dataset))


            test_dataset = np.stack([SpecObj.data[int(lowbound//SpecObj.freq_res)+1:int(highbound//SpecObj.freq_res)+2] for SpecObj in self.data if (SpecObj.source, SpecObj.subject) in test_subjects])
            test_labels = np.array([int(1) if (SpecObj.group==2) else int(-1) for SpecObj in self.data if (SpecObj.source, SpecObj.subject) in test_subjects])
            # print("Number of test examples:", len(test_dataset))

            # print("Frequency Resolution:", self.data[0].freq_res, "Hz")

        print("Number of samples in train:", train_dataset.shape[0])
        print("Number of samples in test:", test_dataset.shape[0])

        # reshape datasets
        nsamples, nx, ny = train_dataset.shape
        train_dataset = train_dataset.reshape((nsamples, nx*ny))

        nsamples, nx, ny = test_dataset.shape
        test_dataset = test_dataset.reshape((nsamples, nx*ny))


        # << Normalize features (z-score standardization) >> #
        # create classifier pipeline
        if normalize != None:
            if kernel == 'linear':
                clf = make_pipeline(myscaler, LinearSVC(C=C, random_state=0, max_iter=iterations, dual=False))
            elif kernel == 'rbf':
                clf = make_pipeline(myscaler, SVC(kernel='rbf', C=C, random_state=0, max_iter=iterations))
        else:
            if kernel == 'linear':
                clf = make_pipeline(LinearSVC(C=C, random_state=0, max_iter=iterations, dual=False))
            elif kernel == 'rbf':
                clf = make_pipeline(SVC(kernel='rbf', C=C, random_state=0, max_iter=iterations))

        # fit to train data
        clf.fit(train_dataset, train_labels)
        print('Classification accuracy on validation data: {:.3f}'.format(clf.score(test_dataset, test_labels)))

        # Compare to the weights of an SVM
        # only available using linear kernel
        if kernel == "linear":
            svm_weights = np.abs(clf[-1].coef_).sum(axis=0)
            svm_weights /= svm_weights.sum()

        if plot_PR==True:
            # test_score = clf.decision_function(test_dataset)
            # average_precision = average_precision_score(test_dataset, test_score)

            # print('Average precision-recall score: {0:0.2f}'.format(average_precision))

            disp = plot_precision_recall_curve(clf, test_dataset, test_labels)
            disp.ax_.set_title('2-class Precision-Recall curve: ' 'AP={0:0.2f}'.format(average_precision))


        # univariate feature selection with F-test for feature scoring
        if feat_select==True:
            # We use the default selection function to select the ten
            # most significant features
            selector = SelectKBest(f_classif, k=num_feats)
            selector.fit(train_dataset, train_labels)
            scores = -np.log10(selector.pvalues_)
            scores /= scores.max()

            # Compare to the selected-weights of an SVM
            if normalize != None:
                if kernel == 'linear':
                    clf_selected = make_pipeline(SelectKBest(f_classif, k=num_feats), myscaler, LinearSVC(C=C, random_state=0, max_iter=iterations, dual=False))
                elif kernel == 'rbf':
                    clf_selected = make_pipeline(SelectKBest(f_classic, k=num_feats), myscaler, SVC(kernel='rbf', C=C, random_state=0, max_iter=iterations))
            else:
                if kernel == 'linear':
                    clf_selected = make_pipeline(SelectKBest(f_classif, k=num_feats), LinearSVC(C=C, random_state=0, max_iter=iterations, dual=False))
                elif kernel == 'rbf':
                    clf_selected = make_pipeline(SelectKbest(f_classic, k=num_feats), SVC(kernel='rbf', C=C, random_state=0, max_iter=iterations))

            clf_selected.fit(train_dataset, train_labels)
            print('Classification accuracy after univariate feature selection: {:.3f}'.format(clf_selected.score(test_dataset, test_labels)))

            # Compare to the weights of an SVM
            # only available using linear kernel
            if kernel == 'linear':
                svm_weights_selected = np.abs(clf_selected[-1].coef_).sum(axis=0)
                svm_weights_selected /= svm_weights_selected.sum()

        if plot_Features==True:
            if kernel == 'linear':
                # set up figure and axes (rows)
                fig, axs = plt.subplots(nrows=len(self.network_channels), figsize=(20, 40))
                plt.rcParams['figure.dpi'] = 200

                # plt.clf()

                # X_indices = np.arange(train_dataset.shape[-1])
                X_indices = Features

                i = 0
                j = 0
                for channel in self.network_channels:
                    axs[j].set_title(channel)
                    axs[j].bar(X_indices - .25, svm_weights[i:i+len(X_indices)], width=.1, label='SVM weight')
                    if feat_select != True:
                        axs[j].legend()
                    i+=len(X_indices)
                    j+=1

                if feat_select==True:
                    i = 0
                    j = 0
                    for channel in self.network_channels:
                        axs[j].bar(X_indices - .45, scores[i:i+len(X_indices)], width=.1, label=r'Univariate score ($-Log(p_{value})$)')
                        # axs[j].bar(X_indices[selector.get_support()] - .05, svm_weights_selected[i:i+len(X_indices)], width=.1, label='SVM weights after selection')
                        axs[j].legend()
                        i+=len(X_indices)
                        j+=1

                axs[0].set_title("Comparing feature selection")
                axs[-1].set_xlabel('Feature number (in Hz)')
                # plt.yticks(())
                fig.tight_layout()
                plt.show()



# ===================== WAVi Analysis Toolbox Config Overview ===================== #
# Welcome to the WAVi EEG Analysis Toolbox. Before running this code, set your      #
# configuration variables below. Most classes, scripts, etc., will reference it.    #

# INITIALIZING STUDY FILES
# ================================================================================= #
# before beginning with new data, format your new study directory as follows:       #
# ~/.../path/to/mystudies/newstudydir
# ----------> /raw
# --------------------> *.eeg
# --------------------> *.evt
# --------------------> *.art

# ----------------------
myStudies = "/wavi/EEGstudies" # if you are working with multiple studies at once, set the parent directory where you will keep them all
studyDirectory = myStudies+"/BrainStudy" # specific study, for functions that only deal with 1 at a time

tasks = {
    "P300": ["p300", "P300_Eyes_Closed", "P300-Sync_Blink", "P300_Eye_Closed", "P300s", "P300-2"],
    "FLNK": ["Flanker", "flanker"],
    "CLSD": ["Eyes_Closed_Resting", "CLSD-2"],
    "OPEN": ["Eyes_Open_Focused"],
    "REST": ["Rest", "rest"],
    "CRNC": ["Chronic", "chronic"],
    "SOMA": ["SMS"]
}


# this package expects a naming convention for raw EEG files:
# n-digit participant number, underscore, task name, .art / .eeg / .evt
# Ex: 104_p300.eeg
# if you want to use a different length participant identifier, specify it here
participantNumLen = 4 # default length
sampleRate = 250 # in Hz
artDegree = 2 # highest number of WAVi-supplied artifact
# still accepted by the program, 0 (strict), 1 (loose), or 2 (none)


# default channel names, customize if using non-WAVi headset
channel_names = [
    'Fp1',
    'Fp2',
    'F3',
    'F4',
    'F7',
    'F8',
    'C3',
    'C4',
    'P3',
    'P4',
    'O1',
    'O2',
    'T3',
    'T4',
    'T5',
    'T6',
    'Fz',
    'Cz',
    'Pz'
]

# channels to be used for artifacting, contigification
# and ultimately used in neural net analyses
# for accurate sensors in spectral analysis,
# keep these in the same order
# as the default list above (channel_names)
network_channels = [
    'Fp1',
    'Fp2',
    'F3',
    'F4',
    'F7',
    'F8',
    'C3',
    'C4',
    'P3',
    'P4',
    'O1',
    'O2',
    'T3',
    'T4',
    'T5',
    'T6',
    'Fz',
    'Cz',
    'Pz'
]

# kernel type for SVM analysis
kernel_type = "linear"
